from agent import Agent, SimpleAgent
from pong import *
import select
from matplotlib import pyplot as plt
import collections

from sample_submission import RandomAgent


class ObservationParser:

    def __init__(self, history=5):
        self.observations = collections.deque(maxlen=history)

    def parse(self, observation):
        self.observations.append(observation)
        return self._features(self.observations)

    def _features(self, observations):
        paddle, other_paddle, ball, table_size = observations[-1]

        f = [paddle.pos[0] / table_size[0], paddle.pos[1] / table_size[1],
             other_paddle.pos[0] / table_size[0], other_paddle.pos[1] / table_size[1]]

        ballx = ball.pos[0] / table_size[0]*2-1
        bally = ball.pos[1] / table_size[1]
        f.append(ballx if paddle.pos[0] > table_size[0]/2 else -ballx)
        f.append(bally)

        speedx = (observations[-1][2].pos[0] - observations[0][2].pos[0]) / table_size[0]
        speedy = (observations[-1][2].pos[1] - observations[0][2].pos[1]) / table_size[1]
        f.append(speedx if paddle.pos[0] > table_size[0]/2 else -speedx)
        f.append(speedy)

        return f

    def __call__(self, observation):
        return self.parse(observation)


def play_epoch(agent1: Agent, agent2: Agent):
    paddles = [Paddle((20, g_table_size[1] / 2), g_paddle_size, g_paddle_speed, g_max_angle, 1, g_timeout),
               Paddle((g_table_size[0] - 20, g_table_size[1] / 2), g_paddle_size, g_paddle_speed, g_max_angle, 0, g_timeout)]
    ball = Ball(g_table_size, g_ball_size, g_paddle_bounce, g_wall_bounce, g_dust_error, g_init_speed_mag)

    p1 = ObservationParser()
    p2 = ObservationParser()
    paddles[0].move_getter = lambda *obs: agent1.action(p1(obs))
    paddles[1].move_getter = lambda *obs: agent2.action(p2(obs))

    score = [0, 0]

    steps = 0
    while max(score) < 1:
        ball, score = check_point(score, ball, g_table_size)
        paddles[0].move(paddles[1].frect, ball.frect, g_table_size)
        paddles[1].move(paddles[0].frect, ball.frect, g_table_size)

        inv_move_factor = int((ball.speed[0] ** 2 + ball.speed[1] ** 2) ** .5)
        if inv_move_factor > 0:
            for i in range(inv_move_factor):
                ball.move(paddles, g_table_size, 1. / inv_move_factor)
        else:
            ball.move(paddles, g_table_size, 1)

        reward = score[0] - score[1]
        agent1.learn(reward)
        agent2.learn(-reward)
        steps += 1

    agent1.epoch_end()
    agent2.epoch_end()

    return score, steps


def evaluate(agent1, agent2, matches=99):
    pygame.init()

    score = [0,0]
    for i in range(matches):
        s, _ = play_epoch(agent1, agent2)
        score[0] += s[0]
        score[1] += s[1]
    print(score)

    pygame.quit()


def evaluate_watch(agent1, agent2):
    pygame.init()
    p1 = ObservationParser()
    p2 = ObservationParser()
    init_game(lambda *obs: agent1.get_apply_agent().action(p1(obs)), lambda *obs: agent2.get_apply_agent().action(p2(obs)), 10)
    pygame.quit()


def train(agent1, agent2, filename=None, epochs=100):
    pygame.init()

    steps_list = []
    avg_steps = []
    for i in range(epochs):
        if i % 2 == 0:
            score, steps = play_epoch(agent1, agent2)
            reward = score[0] - score[1]
        else:
            score, steps = play_epoch(agent2, agent1)
            reward = score[1] - score[0]
        steps_list.append(steps)

        print(i, reward, steps)

        if i != 0 and i % 100 == 0:
            fig = plt.figure(1)
            fig.clear()
            avg_steps.append(sum(steps_list)/len(steps_list))
            plt.plot(avg_steps)
            if filename is not None:
                agent1.save(filename+"-1.model")
                agent1.save(filename+"-2.model")
                plt.savefig(filename+".png")
            plt.show(block=False)
            plt.pause(0.001)

        # if there is an input => show a match
        if select.select([sys.stdin], [], [], 0)[0]:
            while select.select([sys.stdin], [], [], 0)[0]:
                os.read(sys.stdin.fileno(), 4096)
            aa1 = agent1.get_apply_agent()
            aa2 = agent2.get_apply_agent()
            p1 = ObservationParser()
            p2 = ObservationParser()
            init_game(lambda *obs: aa1.action(p1(obs)), lambda *obs: aa2.action(p2(obs)), 3)

    pygame.quit()


def train_selfplay(agent, filename=None, epochs=100):
    train(agent, agent.get_apply_agent(), filename, epochs)


if __name__ == '__main__':
    epochs = 80000
    agent = RandomAgent()

    train_selfplay(agent)

    evaluate(agent.get_apply_agent(), SimpleAgent())
